/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"
#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"
#include "if.h"
#include "sockunion.h"
#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"
#include <string.h>
#include <netinet/igmp.h>
#include <linux/mroute.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "assert.h"
#include "tables.h"
#include "pim_route.h"
#include "pim_packet.h"
#include "igmp_def.h"
#include "ioctl.h"
#include "if_idx_nexthop.h"
#include "pim_assert.h"
#include "pim_joinprune.h"

#include "zbug.h"
ZBUG_EXTERN

extern int pim_sock;
extern int in_cksum (void *ptr, int nbytes);
extern list pim_neighbor_list;
extern list mrt_hash[];
extern list mgrp_hash[];
extern list msrc_hash[MSRC_HASH_SIZE];
extern struct best_route best_route;
extern int pim_sock, igmp_sock;


char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */
char pim_sendbuf[4096];

char *
inet_2a(u_int32_t a, char *b)
{
  sprintf(b, "%u.%u.%u.%u", 
	  (a    ) & 0xFF,
	  (a>> 8) & 0xFF, 
	  (a>>16) & 0xFF, 
	  (a>>24) & 0xFF);
  return  b;
}

int
decode_u_int16(char *p, u_int16_t *v) {
  memcpy(v, &p[0], 2);
  return 2;
}

int
decode_u_int32(char *p, u_int32_t *v) {
  memcpy(v, &p[0], 4);
  return 4;
}

int
decode_unicast(char *p, struct encoded_unicast *eu)
{
  if(p[0] != 1 || p[1] != 0) { 
    printf("Error in codning family/type %d/%d\n",
           p[0], p[1]);
    return 0;
  }
  eu->family=p[0];
  eu->type=p[1];
  memcpy(&eu->addr, &p[2], 4);
  return 6;
}

int
encode_unicast(char *p, u_int32_t addr)
{
  
  p[0] = 1; /* Family */
  p[1] = 0; /* Type */
  memcpy(&p[2], &addr, 4);
  return 6;
}

#define WARN 1
#define NOWARN 0

int
decode_group(char *p, struct encoded_group *eg, int warning)
{
  if(p[0] != 1 || p[1] != 0) { 
    printf("Error in codning family/type %d/%d\n",
           p[0], p[1]);
    return 0;
  }
  eg->family=p[0];
  eg->type=p[1];
  eg->masklen = p[3];

  if(warning && eg->masklen != 32 ) {

    /* We have a problem. The spec says we should accept other
       masks than 32. But currently we have no kernel support
       now. Can we do something later or change the PIM spec?  */
   
    printf("Error. Groups with masks 32 is only supported. Received mask %d\n", 
           eg->masklen);
  }

  memcpy(&eg->addr, &p[4], 4);
//  eg->addr = ntohl(eg->addr);
  return 8;
}

int
encode_group(char *p, u_int32_t addr, u_int8_t masklen)
{
  p[0] = 1; /* Family */
  p[1] = 0; /* Type */
  p[2] = 0; /* Reserved */
  p[3] = masklen;
  memcpy(&p[4], &addr, 4);
  return 8;
}

int
decode_source(char *p, struct encoded_source *es)
{
  if(p[0] != 1 || p[1] != 0) { 
    printf("Error in codning family/type %d/%d\n",
           p[0], p[1]);
    return 0;
  }
  es->family=p[0];
  es->type=p[1];
  es->bits = p[2];
  es->masklen = p[3];

  if(es->masklen != 32 ) {

    /* We have a problem. The spec says we should accept other
       masks than 32. But currently we have no kernel support
       now. Can we do something later or change the PIM spec?  */
   
    printf("Error. Sources with masks 32 is only supported. Received mask %d\n", 
           es->masklen);
    return 0;
  }
  memcpy(&es->addr, &p[4], 4);
  return 8;
}

int
encode_source(char *p, u_int32_t addr, u_int8_t masklen, u_int8_t bits)
{
  
  p[0] = 1; /* Family */
  p[1] = 0; /* Type */
  p[2] = bits; 
  p[3] = masklen;
  memcpy(&p[4], &addr, 4);
  return 8;
}

void
parse_join_prune(struct in_addr *src, char *ph, int len,
		     struct interface *ifp)
{
  char *p =  (ph + sizeof(struct pimheader));
  struct encoded_unicast eu;
  u_int8_t num_groups;
  u_int16_t holdtime;
  int i, ngrps;
  struct in_addr eu_addr;

  /* TODO: Add len checking */

   i = decode_unicast(p, &eu);
   if(i>0) p += i; else goto error;

   /* If our address is the Upstream Neighbor Address its
      for us */

   eu_addr.s_addr = eu.addr;
   if(!local_in_addr((struct in_addr *) &eu.addr)) { // FIXME argument wrong type!!
     printf("Join/Prune for %s is not for us\n", inet_2a(eu.addr, b1));
     return;
   }

   p++; /* reserved */
   num_groups = *p++;

   i = decode_u_int16(p, &holdtime);
   if(i>0) p += i; else goto error;
   holdtime = ntohs(holdtime);

  printf("holdtime %u\n", holdtime);

  for(ngrps=0; ngrps < num_groups; ngrps++) {
    struct encoded_group eg;
    u_int16_t num_joins, num_prunes;
    int j,pr;

    i = decode_group(p, &eg, WARN);
    if(i>0) p += i; else goto error;

    i = decode_u_int16(p, &num_joins);
    if(i>0) p += i; else goto error;
    num_joins = ntohs(num_joins);

    i = decode_u_int16(p, &num_prunes);
    if(i>0) p += i; else goto error;
    num_prunes = ntohs(num_prunes);

    printf("%s/%d Joins/Prunes %d/%d\n", inet_2a( eg.addr, b1), 
	   eg.masklen, num_joins, num_prunes );

    for(j=0; j<num_joins; j++) {
      struct encoded_source es;
      i = decode_source(p, &es);
      if(i>0) p += i; else goto error;


      printf("Join %s %s%s%s \n", inet_2a( es.addr, b1), 
	     es.bits&S_BIT? "S":"",
	     es.bits&WC_BIT? "W":"",
	     es.bits&RPT_BIT? "R":"");


      /* Handle the situations */

      if( es.bits&WC_BIT && es.bits&RPT_BIT && es.bits&S_BIT)
	parse_join_rt_1_wc_1(holdtime, &eg, &es, ifp);

      if( !(es.bits&WC_BIT) && !(es.bits&RPT_BIT) && es.bits&S_BIT)
	parse_join_rt_0_wc_0(holdtime, &eg, &es, ifp);
      
    }

    for(pr=0; pr<num_prunes; pr++) {
      struct encoded_source es;
      i = decode_source(p, &es);
      if(i>0) p += i; else goto error;

      printf("Prune %s %s%s%s \n", inet_2a( es.addr, b1), 
	     es.bits&S_BIT? "":"S",
	     es.bits&WC_BIT? "":"W",
	     es.bits&RPT_BIT? "":"R");


      /* Handle the situations */

      if( es.bits&WC_BIT && es.bits&RPT_BIT && es.bits&S_BIT)
	parse_prune_rt_1_wc_1(holdtime, &eg, &es, ifp);

      if( !(es.bits&WC_BIT) && !(es.bits&RPT_BIT) && es.bits&S_BIT)
	parse_prune_rt_0_wc_0(holdtime, &eg, &es, ifp);

      if( (es.bits&WC_BIT) && !(es.bits&RPT_BIT) && es.bits&S_BIT)
	parse_prune_rt_1_wc_0(holdtime, &eg, &es, ifp);
      
    }
  }
  return;
error:
  return;
}

void
parse_assert(struct in_addr *src, char *ph, int len, struct interface *ifp)
{
  char *p =  (ph + sizeof(struct pimheader));
  struct encoded_group eg;
  struct encoded_unicast eu;
  u_int32_t metric, preference, rpt;
  int i;

  i = decode_group(p, &eg, WARN);
  if(i>0) p += i; else goto error;

  i = decode_unicast(p, &eu);
  if(i>0) p += i; else goto error;

  i = decode_u_int32(p, &preference);
  if(i>0) p += i; else goto error;
  preference = ntohl(preference);

  i = decode_u_int32(p, &metric);
  if(i>0) p += i; else goto error;
  metric = ntohl(metric);

  if(preference & ASSERT_RPT_BIT) rpt = 1;
  else rpt = 0;

  preference &= ~ASSERT_RPT_BIT;

  process_assert(&eg, 
		 &eu, 
		 rpt,
		 preference,
		 metric,
		 ifp,
		 src->s_addr);


  return;
error:
  return;
}

void
parse_register(struct in_addr *src, char *ph, int len, 
	       struct interface *ifp)
{
  char *p =  (ph + sizeof(struct pimheader));
  u_int32_t reserved;
  int i, B_bit, N_bit;;
  
  i = decode_u_int32(p, &reserved);
  if(i>0) p += i; else goto error;

  if(reserved & 0x80000000) B_bit =1;
  if(reserved & 0x40000000) N_bit =1;

  /* Register packet starts at p and is 
     len-28

     20 IP header
      4 PIM Header
      4 Reserved
      */

    ZBUG(ZBUG_PIM_REGISTER_PACKET) 
      {
	ZBUGF("RX REGISTER (%s) ", ifp->name);
	for(i=0; i < len-28; i++) ZBUGF ("%x ", p[i]&0xFF);
	ZBUGF("\n"); 
      }
    return;
error:
    return;
}

void
stop_all_registers()
{
  /* TODO */
}

int
register_stop_timeout(struct thread *t_register_stop)
{
  struct mrt *mr = THREAD_ARG (t_register_stop);
  vifbitmap_t oifs = 0;

  zassert(mr, -1);
  mr->t_register_stop = NULL;
  mr->flags &= ~MRT_REGISTER_STOP;

#if 0
  if(mr->oifs == 0)  {
    mr_remove(mr);
    return 1;
  }
#endif

  oifs = mr_oifs(mr);
  mfc_add(igmp_sock,
	  mr->ms->src,
	  mr->mg->group,
	  mr_iif( mr ), 
//mr->ms->ifp->ifindex, 
	  oifs);
   
  return 1;
}

void
parse_register_stop(struct in_addr *src, char *ph, int len, 
		    struct interface *ifp)
{
  char *p =  (ph + sizeof(struct pimheader));
  struct encoded_group eg;
  struct encoded_unicast eu;
  struct mrt *mr;
  u_int32_t timer;
  vifbitmap_t oifs = 0;
  int i;

  i = decode_group(p, &eg, WARN);
  if(i>0) p += i; else goto error;

  i = decode_unicast(p, &eu);
  if(i>0) p += i; else goto error;

  ZBUG(ZBUG_PIM_REGISTER_PACKET) 
    printf("Register-Stop (S,G) = %s, %s \n", inet_2a(eg.addr, b1), 
	   inet_2a(eu.addr, b2));

  if( eg.addr == 0 ) stop_all_registers(); /* Special case */
  
  mr = mr_lookup(eu.addr, eg.addr);
  if(!mr) {
    ZBUG(ZBUG_PIM_REGISTER_PACKET) 
      printf("Register-Stop error no (S,G) entry for %s %s\n",
	     inet_2a(eu.addr, b1), inet_2a(eg.addr, b2));
    return;
  }
	     
  mr->flags |= MRT_REGISTER_STOP;
  
  timer = REGISTER_STOP_TIMER;

  if(mr->t_register_stop) thread_cancel(mr->t_register_stop);
  mr->t_register_stop = thread_add_timer(master, register_stop_timeout,
					 mr, timer);

  oifs = mr_oifs(mr);
  mfc_add(igmp_sock,
	  mr->ms->src,
	  mr->mg->group,
	  mr_iif( mr ),
//mr->ms->ifp->ifindex, 
	  oifs);
   
  return;
error:
  return;
}

int
hello_thread_timeout(struct thread *t_hello_timeout)
{
  struct pim_neighbor *pn = THREAD_ARG (t_hello_timeout);
  pn->t_hello_neighbor = NULL;
  pim_neighbor_remove(pn);
  return 1;
}

void
parse_hello(struct in_addr *src, char *ph, int len, struct interface *ifp)
{
  struct hello_option *h;
  struct pim_neighbor *pn;
  int type, olen, holdtime;
  h = (struct hello_option *)  (ph + sizeof(struct pimheader));
  
  if(local_in_addr(src)) return;

  type = htons(h->type); /* Holdtime message attributes */
  olen = htons(h->length);
  holdtime = htons(h->value);
  
  if( type != 1 && olen != 2 ) {
    zlog_warn("Expected PIM HELLO got options(TLV %d/%d/%d) from %s",
	      ntohs(h->type), ntohs( h->length), ntohs(h->value), 
	      inet_ntoa((struct in_addr) *src));
    return;
  }
  
  if(0) zlog_warn("PIM HELLO, options(TLV %d/%d/%d) from %s\n",
		  ntohs(h->type), ntohs( h->length), ntohs(h->value), 
		  inet_ntoa((struct in_addr) *src));
 
  /* Look up neighbor or create new and handle neighbor leave */
  pn = pim_neighbor_list_lookup_from_in_addr(src);

  if( ! pn ) {
    pn = pim_neighbor_new(src, holdtime);
    pn->version = (((struct pimheader *)(ph))->version) >>4;
    pn->ifp = ifp;
    select_DR();
    trigger_join_prune(ifp);
  }
  else if( holdtime == 0 ) pim_neighbor_list_delete(pn);


  /* Common path we received a HELLO from an old neigbor */
  else {
      pn->exptime = time (NULL)+holdtime;	   

      /* Check for interface changes and neighbor version changes */

      if( pn->version == (((struct pimheader *)(ph))->version) >>4) 
	 pn->version = (((struct pimheader *)(ph))->version) >>4;
      else {
	/* Handle a version change here */

	zlog_warn("Neighbor is changing PIM version\n");
	pn->version = (((struct pimheader *)(ph))->version) >>4;
      }

      /* Check for interface changes */

      if(pn->ifp != ifp) {
	zlog_warn("Neighbor's  interface has changed!! \n");
	pim_neighbor_remove(pn);
	select_DR();
	return;
      }

      /* Cancel the old timer and initiate a new */
      /* BUT 0xffff should not time out trh neighbor */
      
      if(pn->t_hello_neighbor) {
	thread_cancel(pn->t_hello_neighbor);
	pn->t_hello_neighbor = NULL;
	if(holdtime != 0xffff) 
	  pn->t_hello_neighbor = thread_add_timer(master, 
		    hello_thread_timeout,  pn, holdtime);
      }
  }
  ZBUG(ZBUG_PIM_NEIGHBOR) zlog_warn("RX Hello from %s Holdtime %d", 
			    inet_ntoa((struct in_addr) *src), holdtime);
  
  return;
}

u_int32_t rp_hash_value( u_int32_t group, u_int32_t hash_mask, u_int32_t rp )
{
  return ((uint)1103515245 * (((uint)1103515245 * ( group & hash_mask ) +12345) ^ rp )
	  +12345) & ((uint)(1 << 31) -1) ;
  /* mod 2^31 */
}

void parse_bootstrap(struct in_addr *src,
		     char *ph, int len, struct interface *ifp)
{
  int i;
  char *stop, *p =  (ph + sizeof(struct pimheader));

  u_int8_t hash_mask_len, BSR_priority, RP_count, frag_RP_count, priority;
  u_int16_t fragment_tag, holdtime;
  struct encoded_unicast BSR, RP;
  struct encoded_group group;

  /* Check that the message comes from our RPF-neighbor */

  if_idx_nexthop(src);
  if( get_iflist_ifp(best_route.idx) != ifp) return;

  len -= sizeof(struct pimheader) ;
  stop = ph + len;

  i = decode_u_int16(p, &fragment_tag);
  if(i>0) p += i; else goto error;
  fragment_tag = ntohs(fragment_tag);

  hash_mask_len = *p++;
  BSR_priority = *p++;

  i = decode_unicast(p, &BSR);
  if(i>0) p += i; else goto error;
  
  printf("BOOTSTRAP from bsr %s priority %u mask %u\n",
	 inet_2a(BSR.addr, b1),
	 BSR_priority,
	 hash_mask_len);

  while( p < stop )
    {
      i = decode_group(p, &group, NOWARN);
      if(i>0) p += i; else goto error;
  
      RP_count = *p++;
      frag_RP_count = *p++;

      printf(" group %s/%d count %u frag %u\n",
	     inet_2a(group.addr, b1),
	     group.masklen,
	     RP_count,
	     frag_RP_count);

      p += 2; /* Reserved */

      while( frag_RP_count-- )
	{
	  i = decode_unicast(p, &RP);
	  if(i>0) p += i; else goto error;
      
	  i = decode_u_int16(p, &holdtime);
	  if(i>0) p += i; else goto error;
	  holdtime = ntohs(holdtime);

	  priority = *p++;
	  p++; /* Reserved */

	  printf(" rp %s priority %u holdtime %u\n",
		 inet_2a(RP.addr, b1),
		 priority,
		 holdtime);
	}
    } 
      
  /* forward bootstrap on all pim_interfaces */
      
      return;
error:
      return;
}

int
pim_neighbor_remove(struct pim_neighbor *pn)
{
  ZBUG(ZBUG_PIM_NEIGHBOR)
     zlog_warn("Removing neighbor %s",inet_ntoa((struct in_addr) pn->address));
  
  /* Clear up pending timers */

  if(pn->t_hello_neighbor) {
    thread_cancel(pn->t_hello_neighbor);
    pn->t_hello_neighbor = NULL;
  }
  pim_neighbor_list_delete(pn);
  XFREE(MTYPE_PIM_NEIGHBOR, pn);
  return 1;
}

/* Remove all neighbors connected to an interface */

int
pim_all_neighbor_remove(struct interface *ifp)
{
  struct pim_neighbor *p;
  listnode node;
  
  for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
    p = getdata (node);
    if( ifp == p->ifp) {
      pim_neighbor_remove(p);
      return 1;
    }
  }
  return 0;
}

struct pim_neighbor *
pim_neighbor_new( struct in_addr *src, int holdtime )
{
  struct pim_neighbor *pn;

  if(holdtime == 0) return NULL;

  ZBUG(ZBUG_PIM_NEIGHBOR)
     zlog_warn("New neighbor %s",inet_ntoa((struct in_addr) *src) );
   pn = pim_neighbor_alloc ();
   pim_neighbor_list_add(pn);
   pn->address = *src;
   pn->uptime = time (NULL);
   pn->exptime = time (NULL)+holdtime;

   /* 0xffff should not time out neighbors */
   if(holdtime != 0xffff )
     pn->t_hello_neighbor = thread_add_timer(master, hello_thread_timeout, 
					pn, holdtime);
   return pn;
}

void 
parse_pim_v2_packet(char *p, int len, struct interface *ifp)
{
  struct ip *ip;
  struct pimheader *pim;
  struct in_addr src;
  struct pim_neighbor *pn;
  int ip_hlen, ip_len, pimtype, pimversion;

  ip = (struct ip *)p ;
 
  ip_hlen = ip->ip_hl*4; 
  ip_len = ntohs(ip->ip_len);
  len = len - ip_hlen;

  src = ip->ip_src;

  /* TODO: CHECKSUM TEST & PROTOCOL TEST*/

  /* TODO: Check IP options */


  ZBUG(ZBUG_PIM_PACKET)
    zlog_warn("ip_len %d, proto = %d\n", ip_len, ip->ip_p);
  pim = (struct pimheader *) (p+ip_hlen);

  pimtype =  pim->version & 0xF;
  pimversion =  (pim->version>>4) & 0xFF;

  ZBUG(ZBUG_PIM_PACKET)
    zlog_warn("pim version = %x ", (pim->version>>4) & 0xFF  );
  
  pn = pim_neighbor_list_lookup_from_in_addr(&src);

  switch (pimtype) {
    
  case PIM_HELLO:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_HELLO from %s\n", inet_ntoa(src));
    parse_hello(&src, (char *) pim, len, ifp);
    break;
    
  case PIM_REGISTER:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("PIM_REGISTER from %s\n", inet_ntoa(src));
    parse_register(&src, (char *) pim, len, ifp);
    break;

  case PIM_REGISTER_STOP:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_REGISTER_STOP from %s\n", inet_ntoa(src));
    parse_register_stop(&src, (char *) pim, len, ifp);
    break;

  case PIM_JOIN_PRUNE:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_JOIN_PRUNE from %s\n", inet_ntoa(src)); 
    if(pn) parse_join_prune(&src, (char *) pim, len, ifp);
    break;

  case PIM_BOOTSTRAP:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_BOOTSTAP from %s\n", inet_ntoa(src));
    parse_bootstrap(&src, (char *) pim, len, ifp); 
    break;

  case PIM_ASSERT:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_ASSERT from %s\n", inet_ntoa(src)); 
    if(pn) parse_assert(&src, (char *) pim, len, ifp);
    break;

  case PIM_GRAFT:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_GRAFT from %s\n", inet_ntoa(src)); 
    break;

  case PIM_GRAFT_ACK:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_GRAFT_ACK from %s\n", inet_ntoa(src)); 
    break;

  case PIM_CANDIDATE_RP_ADVERTISEMENT:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX PIM_CANDIDATE_RP_ADVERTISEMEN from %s\n", inet_ntoa(src)); 
    break;

  default:
    ZBUG(ZBUG_PIM_MESSAGES) 
      printf("RX Error Illegal PIM message %d from %s\n", pimtype, 
	     inet_ntoa(src));

  }
}


int
pim_recvmsg (int sock, u_char *buf, int size, int *ifindex)
{
  struct msghdr msg;
  struct iovec iov;
  struct cmsghdr *ptr;
  char adata[1024];
  int ret;

  msg.msg_name = (void *)0;
  msg.msg_namelen = 0;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = (void *) adata;
  msg.msg_controllen = sizeof adata;

  iov.iov_base = buf;
  iov.iov_len = size;

  ret = recvmsg (sock, &msg, 0);
  if (ret < 0) {
    zlog_warn("recvmsg: read error %", strerror(errno));
    return ret;
  }

  if (msg.msg_flags & MSG_TRUNC) {
    zlog_warn("recvmsg: truncated message");
    return ret;
  }
  if (msg.msg_flags & MSG_CTRUNC) {
    zlog_warn("recvmsg: truncated control message");
    return ret;
  }

  for (ptr = CMSG_FIRSTHDR(&msg); ptr ; ptr = CMSG_NXTHDR(&msg, ptr))
    if (ptr->cmsg_level == SOL_IP && ptr->cmsg_type == IP_PKTINFO) 
      {
        struct in_pktinfo *pktinfo;
           pktinfo = (struct in_pktinfo *) CMSG_DATA (ptr);
        *ifindex = pktinfo->ipi_ifindex;
      }
  return ret;
}


int pim_read_raw(struct thread *r)
{
  struct interface *ifp;
  char buf[1500];
  int ret, ifindex;
  
  int sock = THREAD_FD (r);
  t_pim_raw = thread_add_read (master, pim_read_raw, NULL, sock);
  
  ret = pim_recvmsg (sock, buf, 1500,  &ifindex);
  
  if (ret < 0) printf ("Error = %d", ret);
  else {
    int i;
    ZBUG(ZBUG_PIM_PACKET) 
      {
	ZBUGF("RX PIM (idx %d) ", ifindex);
	for(i=0; i < ret; i++) ZBUGF ( "%x ", buf[i]&0xFF);
	ZBUGF("\n");
      }
  }
  ifp = get_iflist_ifp(ifindex);
  if( ifp == NULL) return 0;
  parse_pim_v2_packet(buf, ret, ifp);
  return ret;
}

void
send_pim_hello(struct interface *ifp, int holdtime)
{
 struct pim_interface *pi = ifp->info;
 struct hello_option *pimopt;

 pimopt = (struct hello_option *) (pim_sendbuf + sizeof(struct ip) 
   + sizeof(struct pimheader));

 pimopt->type = htons(1); /* Holdtime */
 pimopt->length = htons(2);
 pimopt->value = htons(holdtime);

 if(pi->version == 2)
 send_pim_v2_packet(ifp, pim_sendbuf, sizeof(struct hello_option), 
		    PIM_HELLO, htonl(INADDR_PIM_ALLRTRS_GROUP), 1);

}


void
send_pim_v2_packet(struct interface *ifp, char *buf, int len, int pim_message,
		   u_int32_t dst, u_int32_t ttl)
{
 static struct sockaddr_in sockdst = {AF_INET};
 struct pim_interface *pi;
 struct ip *ip;
 struct pimheader *pim;
 struct msghdr *msg;
 struct cmsghdr *cmsg;
 struct iovec iovector;
 char msgbuf[256];
 struct in_pktinfo *pktinfo;
 u_long src;
 char b;
 
 if (!ifp->flags & IFF_UP) return;

 src = 0; /* Is filled in */

 ip = (struct ip *) buf;
 ip->ip_hl = sizeof(struct ip) >> 2;
 ip->ip_v = IPVERSION;
 ip->ip_tos = 0xC0;
 ip->ip_off = 0L;
 ip->ip_p = IPPROTO_PIM;
 ip->ip_ttl = ttl;
 ip->ip_src.s_addr = src;
 ip->ip_dst.s_addr = dst;

 pim = (struct pimheader *) (buf + 20);
 pi=ifp->info;

 /* Fix for some routers */

 if( PIM_REGISTER_WHOLEPACKET_CRC == pim_message )
   pim->version = ((pi->version)<<4) + (PIM_REGISTER & 0xFF);
 else
  pim->version = ((pi->version)<<4) + (pim_message & 0xFF);


 pim->reserved = 0;
 pim->checksum = 0;

 ip->ip_len  = sizeof(struct ip) + sizeof(struct pimheader) + len;

 b = 0; 
 if( setsockopt(pim_sock,IPPROTO_IP, IP_MULTICAST_LOOP, 
		  (char *)&b,sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));
 
 b = 1;
 if (setsockopt(pim_sock, IPPROTO_IP, IP_HDRINCL,
		(char *) &b, sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));

 b = 0;
 if (setsockopt(pim_sock, IPPROTO_IP, SO_BROADCAST,
		(char *) &b, sizeof(b)) < 0)
   zlog_warn("sendto %s", strerror (errno));
 

 /* For PIM_REGISTER the checksum is to calculated on the pimheader 
    only + reserved is 4 bytes */


 if( pim_message == PIM_REGISTER )
   pim->checksum = in_cksum((u_short *) pim, sizeof(struct pimheader)+4);
 else
   pim->checksum = in_cksum((u_short *) pim, sizeof(struct pimheader)+len);


 bzero(&sockdst,sizeof(sockdst));
 sockdst.sin_family=AF_INET;
 sockdst.sin_addr.s_addr = dst;

 cmsg = (struct cmsghdr *) (msgbuf + sizeof(struct msghdr));
 cmsg->cmsg_len = sizeof(struct cmsghdr) + sizeof(struct in_pktinfo);
 cmsg->cmsg_level = SOL_IP;
 cmsg->cmsg_type = IP_PKTINFO;
 pktinfo = (struct in_pktinfo *) CMSG_DATA(cmsg);
 pktinfo->ipi_ifindex = ifp->ifindex;
 pktinfo->ipi_spec_dst.s_addr = src;
 pktinfo->ipi_addr.s_addr = src;
 
 iovector.iov_base = (void *) buf;
 iovector.iov_len = ip->ip_len; 
 msg = (struct msghdr *) msgbuf;
 msg->msg_name = &sockdst;
 msg->msg_namelen = sizeof(sockdst);
 msg->msg_iov = &iovector;
 msg->msg_iovlen = 1;
 msg->msg_control = cmsg;
 msg->msg_controllen = cmsg->cmsg_len;
 
 if (sendmsg(pim_sock, msg, 0) < 0) {
   zlog_warn("sendto %s", strerror (errno));
 }
 /*   printf("TX on %s idx %d\n", ifp->name, ifp->ifindex); */
}


 







